@@ -1,12 +1,18 @@
+#![doc = include_str!("../README.md")]
+#![forbid(unsafe_code)]
+#![deny(clippy::all)]
+
 #[macro_use]
 mod object;
 mod datetime;
 pub use crate::object::{Dictionary, Object, ObjectId, Stream, StringFormat};
 
 mod document;
+mod incremental_document;
 mod object_stream;
-mod xref;
+pub mod xref;
 pub use crate::document::Document;
+pub use crate::incremental_document::IncrementalDocument;
 
 mod bookmarks;
 pub use crate::bookmarks::Bookmark;
